@@ -1,7 +1,7 @@
 import re
 
 __package__ = 'python-intervals'
-__version__ = '1.7.0'
+__version__ = '1.8.0'
 __licence__ = 'LGPL3'
 __author__ = 'Alexandre Decan'
 __url__ = 'https://github.com/AlexandreDecan/python-intervals'
@@ -16,7 +16,16 @@ __all__ = [
 ]
 
 
-class _PInf:
+class _Singleton():
+    __instance = None
+
+    def __new__(cls, *args, **kwargs):
+        if not cls.__instance:
+            cls.__instance = super(_Singleton, cls).__new__(cls, *args, **kwargs)
+        return cls.__instance
+
+
+class _PInf(_Singleton):
     """
     Represent positive infinity.
     """
@@ -38,7 +47,7 @@ class _PInf:
     def __repr__(self): return '+inf'
 
 
-class _NInf:
+class _NInf(_Singleton):
     """
     Represent negative infinity.
     """
@@ -107,10 +116,12 @@ def empty():
     """
     Create an empty set.
     """
-    return Interval(AtomicInterval(OPEN, inf, -inf, OPEN))
+    if not hasattr(empty, '_instance'):
+        empty._instance = Interval(AtomicInterval(OPEN, inf, -inf, OPEN))
+    return empty._instance
 
 
-def from_string(string, conv, bound=r'.+?', disj=r' ?\| ?', sep=r', ?', left_open=r'\(', 
+def from_string(string, conv, bound=r'.+?', disj=r' ?\| ?', sep=r', ?', left_open=r'\(',
                 left_closed=r'\[', right_open=r'\)', right_closed=r'\]', pinf=r'\+inf', ninf=r'-inf'):
     """
     Parse given string and create an Interval instance.
@@ -168,7 +179,7 @@ def from_string(string, conv, bound=r'.+?', disj=r' ?\| ?', sep=r', ?', left_ope
     return Interval(*intervals)
 
 
-def to_string(interval, conv=repr, disj=' | ', sep=',', left_open='(', 
+def to_string(interval, conv=repr, disj=' | ', sep=',', left_open='(',
               left_closed='[', right_open=')', right_closed=']', pinf='+inf', ninf='-inf'):
     """
     Export given interval (or atomic interval) to string.
@@ -214,10 +225,9 @@ def to_string(interval, conv=repr, disj=' | ', sep=',', left_open='(',
     return disj.join(exported_intervals)
 
 
-
 def from_data(data, conv=None, pinf=float('inf'), ninf=float('-inf')):
     """
-    Import an interval from a piece of data. 
+    Import an interval from a piece of data.
 
     :param data: a list of 4-uples (left, lower, upper, right).
     :param conv: function that is used to convert "lower" and "upper" to bounds, default to identity.
@@ -236,10 +246,10 @@ def from_data(data, conv=None, pinf=float('inf'), ninf=float('-inf')):
         else:
             return conv(bound)
 
-    for item in data: 
+    for item in data:
         left, lower, upper, right = item
         intervals.append(AtomicInterval(
-            left, 
+            left,
             _convert(lower),
             _convert(upper),
             right
@@ -250,13 +260,13 @@ def from_data(data, conv=None, pinf=float('inf'), ninf=float('-inf')):
 def to_data(interval, conv=None, pinf=float('inf'), ninf=float('-inf')):
     """
     Export given interval (or atomic interval) to a list of 4-uples (left, lower,
-    upper, right). 
+    upper, right).
 
     :param interval: an Interval or AtomicInterval instance.
     :param conv: function that convert bounds to "lower" and "upper", default to identity.
     :param pinf: value used to encode positive infinity.
     :param ninf: value used to encode negative infinity.
-    :return: 
+    :return: a list of 4-uples (left, lower, upper, right)
     """
     interval = Interval(interval) if isinstance(interval, AtomicInterval) else interval
     conv = (lambda v: v) if conv is None else conv
@@ -273,9 +283,9 @@ def to_data(interval, conv=None, pinf=float('inf'), ninf=float('-inf')):
 
     for item in interval:
         data.append((
-            item.left, 
-            _convert(item.lower), 
-            _convert(item.upper), 
+            item.left,
+            _convert(item.lower),
+            _convert(item.upper),
             item.right
         ))
     return data
@@ -285,7 +295,8 @@ class AtomicInterval:
     """
     This class represents an atomic interval.
 
-    An atomic interval is a single interval, with a lower and upper bounds, and two (closed or open) boundaries.
+    An atomic interval is a single interval, with a lower and upper bounds,
+    and two (closed or open) boundaries.
     """
 
     __slots__ = ('_left', '_lower', '_upper', '_right')
@@ -294,7 +305,8 @@ class AtomicInterval:
         """
         Create an atomic interval.
 
-        If a bound is set to infinity (regardless of its sign), the corresponding boundary will be exclusive.
+        If a bound is set to infinity (regardless of its sign), the corresponding boundary will
+        be exclusive.
 
         :param left: Boolean indicating whether the left boundary is inclusive (True) or exclusive (False).
         :param lower: lower bound value.
@@ -351,6 +363,43 @@ class AtomicInterval:
             (self._lower == self._upper and (self._left == OPEN or self._right == OPEN))
         )
 
+    def replace(self, left=None, lower=None, upper=None, right=None, ignore_inf=True):
+        """
+        Create a new interval based on the current one and the provided values.
+
+        Callable can be passed instead of values. In that case, it is called with the current
+        corresponding value except if ignore_inf if set (default) and the corresponding
+        bound is an infinity.
+
+        :param left: (a function of) left boundary.
+        :param lower: (a function of) value of the lower bound.
+        :param upper: (a function of) value of the upper bound.
+        :param right: (a function of) right boundary.
+        :param ignore_inf: ignore infinities if functions are provided (default is True).
+        :return: an Interval instance
+        """
+        if callable(left):
+            left = left(self._left)
+        else:
+            left = self._left if left is None else left
+
+        if callable(lower):
+            lower = self._lower if ignore_inf and self._lower in [-inf, inf] else lower(self._lower)
+        else:
+            lower = self._lower if lower is None else lower
+
+        if callable(upper):
+            upper = self._upper if ignore_inf and self._upper in [-inf, inf] else upper(self._upper)
+        else:
+            upper = self._upper if upper is None else upper
+
+        if callable(right):
+            right = right(self._right)
+        else:
+            right = self._right if right is None else right
+
+        return AtomicInterval(left, lower, upper, right)
+
     def overlaps(self, other, permissive=False):
         """
         Test if intervals have any overlapping value.
@@ -520,52 +569,48 @@ class AtomicInterval:
         return not self == other  # Required for Python 2
 
     def __lt__(self, other):
-        if isinstance(other, Interval):
-            other = other.to_atomic()
-
         if isinstance(other, AtomicInterval):
             if self._right == OPEN:
                 return self._upper <= other._lower
             else:
                 return self._upper < other._lower or (self._upper == other._lower and other._left == OPEN)
+        elif isinstance(other, Interval):
+            return self < other.to_atomic()
         else:
-            return NotImplemented
+            return AtomicInterval(CLOSED, other, other, CLOSED) > self
 
     def __gt__(self, other):
-        if isinstance(other, Interval):
-            other = other.to_atomic()
-
         if isinstance(other, AtomicInterval):
             if self._left == OPEN:
                 return self._lower >= other._upper
             else:
                 return self._lower > other._upper or (self._lower == other._upper and other._right == OPEN)
+        elif isinstance(other, Interval):
+            return self > other.to_atomic()
         else:
-            return NotImplemented
+            return AtomicInterval(CLOSED, other, other, CLOSED) < self
 
     def __le__(self, other):
-        if isinstance(other, Interval):
-            other = other.to_atomic()
-
         if isinstance(other, AtomicInterval):
             if self._right == OPEN:
                 return self._upper <= other._upper
             else:
                 return self._upper < other._upper or (self._upper == other._upper and other._right == CLOSED)
+        elif isinstance(other, Interval):
+            return self <= other.to_atomic()
         else:
-            return NotImplemented
+            return AtomicInterval(CLOSED, other, other, CLOSED) >= self
 
     def __ge__(self, other):
-        if isinstance(other, Interval):
-            other = other.to_atomic()
-
         if isinstance(other, AtomicInterval):
             if self._left == OPEN:
                 return self._lower >= other._lower
             else:
                 return self._lower > other._lower or (self._lower == other._lower and other._left == CLOSED)
+        elif isinstance(other, Interval):
+            return self >= other.to_atomic()
         else:
-            return NotImplemented
+            return AtomicInterval(CLOSED, other, other, CLOSED) <= self
 
     def __hash__(self):
         try:
@@ -638,6 +683,34 @@ class Interval:
                 else:
                     i = i + 1
 
+    @property
+    def left(self):
+        """
+        Boolean indicating whether the lowest left boundary is inclusive (True) or exclusive (False).
+        """
+        return self._intervals[0].left
+
+    @property
+    def lower(self):
+        """
+        Lowest lower bound value.
+        """
+        return self._intervals[0].lower
+
+    @property
+    def upper(self):
+        """
+        Highest upper bound value.
+        """
+        return self._intervals[-1].upper
+
+    @property
+    def right(self):
+        """
+        Boolean indicating whether the highest right boundary is inclusive (True) or exclusive (False).
+        """
+        return self._intervals[-1].right
+
     def is_empty(self):
         """
         Test interval emptiness.
@@ -676,6 +749,79 @@ class Interval:
         """
         return Interval(self.to_atomic())
 
+    def replace(self, left=None, lower=None, upper=None, right=None, ignore_inf=True):
+        """
+        Create a new interval based on the current one and the provided values.
+
+        If current interval is not atomic, it is extended or restricted such that
+        its enclosure satisfies the new bounds. In other words, its new enclosure
+        will be equal to self.to_atomic().replace(left, lower, upper, right).
+
+        Callable can be passed instead of values. In that case, it is called with the current
+        corresponding value except if ignore_inf if set (default) and the corresponding
+        bound is an infinity.
+
+        :param left: (a function of) left boundary.
+        :param lower: (a function of) value of the lower bound.
+        :param upper: (a function of) value of the upper bound.
+        :param right: (a function of) right boundary.
+        :param ignore_inf: ignore infinities if functions are provided (default is True).
+        :return: an Interval instance
+        """
+        enclosure = self.to_atomic()
+
+        if callable(left):
+            left = left(enclosure._left)
+        else:
+            left = enclosure._left if left is None else left
+
+        if callable(lower):
+            lower = enclosure._lower if ignore_inf and enclosure._lower in [-inf, inf] else lower(enclosure._lower)
+        else:
+            lower = enclosure._lower if lower is None else lower
+
+        if callable(upper):
+            upper = enclosure._upper if ignore_inf and enclosure._upper in [-inf, inf] else upper(enclosure._upper)
+        else:
+            upper = enclosure._upper if upper is None else upper
+
+        if callable(right):
+            right = right(enclosure._right)
+        else:
+            right = enclosure._right if right is None else right
+
+        n_interval = self & AtomicInterval(left, lower, upper, right)
+
+        if len(n_interval) > 1:
+            lowest = n_interval[0].replace(left=left, lower=lower)
+            highest = n_interval[-1].replace(upper=upper, right=right)
+            return Interval(*[lowest] + n_interval[1:-1] + [highest])
+        else:
+            return Interval(n_interval[0].replace(left, lower, upper, right))
+
+    def apply(self, func):
+        """
+        Apply given function on each of the underlying AtomicInterval instances and return a new
+        Interval instance. The function is expected to return an AtomicInterval, an Interval
+        or a 4-uple (left, lower, upper, right).
+
+        :param func: function to apply on each of the underlying AtomicInterval instances.
+        :return: an Interval instance.
+        """
+        intervals = []
+
+        for interval in self:
+            value = func(interval)
+
+            if isinstance(value, (Interval, AtomicInterval)):
+                intervals.append(value)
+            elif isinstance(value, tuple):
+                intervals.append(AtomicInterval(*value))
+            else:
+                raise TypeError('Unsupported return type {} for {}'.format(type(value), value))
+
+        return Interval(*intervals)
+
     def overlaps(self, other, permissive=False):
         """
         Test if intervals have any overlapping value.
@@ -746,22 +892,13 @@ class Interval:
         return self - other
 
     def __len__(self):
-        if self._intervals[0].is_empty():
-            return 0
-        else:
-            return len(self._intervals)
+        return len(self._intervals)
 
     def __iter__(self):
-        if self._intervals[0].is_empty():
-            return iter([])
-        else:
-            return iter(self._intervals)
+        return iter(self._intervals)
 
     def __getitem__(self, item):
-        if self._intervals[0].is_empty():
-            raise IndexError('Interval is empty')
-        else:
-            return self._intervals[item]
+        return self._intervals[item]
 
     def __and__(self, other):
         if isinstance(other, (AtomicInterval, Interval)):
@@ -836,36 +973,16 @@ class Interval:
         return not self == other  # Required for Python 2
 
     def __lt__(self, other):
-        if isinstance(other, AtomicInterval):
-            return self._intervals[-1] < other
-        elif isinstance(other, Interval):
-            return self._intervals[-1] < other._intervals[0]
-        else:
-            return NotImplemented
+        return self.to_atomic() < other
 
     def __gt__(self, other):
-        if isinstance(other, AtomicInterval):
-            return self._intervals[0] > other
-        elif isinstance(other, Interval):
-            return self._intervals[0] > other._intervals[-1]
-        else:
-            return NotImplemented
+        return self.to_atomic() > other
 
     def __le__(self, other):
-        if isinstance(other, AtomicInterval):
-            return self._intervals[-1] <= other
-        elif isinstance(other, Interval):
-            return self._intervals[-1] <= other._intervals[-1]
-        else:
-            return NotImplemented
+        return self.to_atomic() <= other
 
     def __ge__(self, other):
-        if isinstance(other, AtomicInterval):
-            return self._intervals[0] >= other
-        elif isinstance(other, Interval):
-            return self._intervals[0] >= other._intervals[0]
-        else:
-            return NotImplemented
+        return self.to_atomic() >= other
 
     def __hash__(self):
         return hash(self._intervals[0])
