@@ -21,7 +21,20 @@ def test_infinity_comparisons_with_self(inf):
     assert not (inf != inf)
 
 
-@pytest.mark.parametrize('o', [0, 1, 1.0, 'a', list(), tuple(), dict(), I.closed(0, 1)])
+@pytest.mark.skipif(sys.version_info < (3,0), reason='Not supported in Python 2')
+def test_infinities_are_singletons():
+    assert I._PInf() == I._PInf()
+    assert I._PInf() is I._PInf()
+    assert I._PInf() is -I._NInf()
+
+    assert I._NInf() == I._NInf()
+    assert I._NInf() is I._NInf()
+
+    assert I._PInf() is not I._NInf()
+    assert I._PInf() is not -I._PInf()
+
+
+@pytest.mark.parametrize('o', [0, 1, 1.0, 'a', list(), tuple(), dict()])
 def test_infinity_comparisons_with_objects(o):
     assert o != I.inf and not (o == I.inf)
     assert o < I.inf and I.inf > o
@@ -45,6 +58,9 @@ def test_creation():
     assert I.closed(3, -3) == I.empty()
     assert I.openclosed(3, 3) == I.empty()
 
+    # I.empty() is a singleton
+    assert I.empty() is I.empty()
+
     assert I.Interval(I.closed(0, 1).to_atomic()) == I.closed(0, 1)
     assert I.Interval(I.closed(0, 1)) == I.closed(0, 1)
     assert I.Interval(I.closed(0, 1).to_atomic(), I.closed(2, 3)) == I.closed(0, 1) | I.closed(2, 3)
@@ -54,6 +70,40 @@ def test_creation():
         I.Interval(1)
 
 
+def test_atomic_bounds():
+    i = I.openclosed(1, 2).to_atomic()
+    assert i.left == I.OPEN
+    assert i.right == I.CLOSED
+    assert i.lower == 1
+    assert i.upper == 2
+
+    i = I.openclosed(10, -10).to_atomic()
+    assert i.left == I.OPEN
+    assert i.right == I.OPEN
+    assert i.lower == I.inf
+    assert i.upper == -I.inf
+
+
+def test_bounds():
+    i = I.openclosed(1, 2)
+    assert i.left == I.OPEN
+    assert i.right == I.CLOSED
+    assert i.lower == 1
+    assert i.upper == 2
+
+    i = I.openclosed(10, -10)
+    assert i.left == I.OPEN
+    assert i.right == I.OPEN
+    assert i.lower == I.inf
+    assert i.upper == -I.inf
+
+    i = I.open(0, 1) | I.closed(3, 4)
+    assert i.left == I.OPEN
+    assert i.right == I.CLOSED
+    assert i.lower == 0
+    assert i.upper == 4
+
+
 def test_hash():
     assert hash(I.closed(0, 1).to_atomic()) is not None
     assert hash(I.closed(0, 1)) is not None
@@ -200,6 +250,54 @@ def test_interval_to_atomic():
     assert I.empty().to_atomic() == I.AtomicInterval(False, I.inf, -I.inf, False)
 
 
+def test_replace():
+    i = I.closed(0, 1).to_atomic()
+    assert i.replace() == i
+    assert i.replace(I.OPEN, 2, 3, I.OPEN) == I.open(2, 3)
+    assert i.replace(upper=2, left=I.OPEN) == I.openclosed(0, 2)
+
+    assert i.replace(lower=lambda v: 1 + v) == I.singleton(1)
+    assert i.replace(left=lambda v: not v, right=lambda v: not v) == I.open(0, 1)
+
+    i = I.open(-I.inf, I.inf)
+    assert i.replace(lower=lambda v: 1, upper=lambda v: 1) == I.open(-I.inf, I.inf)
+    assert i.replace(lower=lambda v: 1, upper=lambda v: 2, ignore_inf=False) == I.open(1, 2)
+    
+    i = I.closed(0, 1) | I.open(2, 3)
+    assert i.replace() == i
+    assert i.replace(I.OPEN, -1, 4, I.OPEN) == I.openclosed(-1, 1) | I.open(2, 4)
+    assert i.replace(lower=2) == I.closedopen(2, 3)
+    assert i.replace(upper=1) == I.closedopen(0, 1)
+    assert i.replace(lower=5) == I.empty()
+    assert i.replace(upper=-5) == I.empty()
+    assert i.replace(left=lambda v: not v, lower=lambda v: v - 1, upper=lambda v: v + 1, right=lambda v: not v) == I.openclosed(-1, 1) | I.openclosed(2, 4)
+
+    assert I.empty().replace(lower=2, upper=4) == I.open(2, 4)
+
+
+def test_apply():
+    i = I.closed(0, 1)
+    assert i.apply(lambda s: s) == i
+    assert i.apply(lambda s: (False, -1, 2, False)) == I.open(-1, 2)
+    assert i.apply(lambda s: I.AtomicInterval(False, -1, 2, False)) == I.open(-1, 2)
+    assert i.apply(lambda s: I.open(-1, 2)) == I.open(-1, 2)
+
+    i = I.closed(0, 1) | I.closed(2, 3)
+    assert i.apply(lambda s: s) == i
+    assert i.apply(lambda s: (False, -1, 2, False)) == I.open(-1, 2)
+    assert i.apply(lambda s: (not s.left, s.lower - 1, s.upper - 1, not s.right)) == I.open(-1, 0) | I.open(1, 2)
+    assert i.apply(lambda s: I.AtomicInterval(False, -1, 2, False)) == I.open(-1, 2)
+    assert i.apply(lambda s: I.open(-1, 2)) == I.open(-1, 2)
+
+    assert i.apply(lambda s: (s.left, s.lower, s.upper * 2, s.right)) == I.closed(0, 6)
+
+    with pytest.raises(TypeError):
+        i.apply(lambda s: None)
+
+    with pytest.raises(TypeError):
+        i.apply(lambda s: 'unsupported')
+        
+
 def test_overlaps():
     # Overlaps should reject non supported types
     with pytest.raises(TypeError):
@@ -293,17 +391,39 @@ def test_atomic_comparisons(i1, i2, i3):
     assert not i1 == 1
 
 
-@pytest.mark.parametrize('i1', [I.closed(0, 1).to_atomic(), I.closed(0, 1)])
-@pytest.mark.xfail(sys.version_info < (3, 0), reason='Python 2 does not raise TypeError for unsupported comparisons')
-def test_comparisons_for_unsupported_types(i1):
-    with pytest.raises(TypeError):
-        i1 > 1
-    with pytest.raises(TypeError):
-        i1 >= 1
-    with pytest.raises(TypeError):
-        i1 < 1
-    with pytest.raises(TypeError):
-        i1 <= 1
+@pytest.mark.parametrize('i', [I.closedopen(0, 10).to_atomic(), I.closedopen(0, 10)])
+def test_comparisons_with_values(i):
+    assert -1 < i
+    assert -1 <= i
+    assert not (0 < i) 
+    assert 0 <= i
+    assert not (5 < i) 
+    assert 5 <= i
+    assert not (10 < i) 
+    assert not (10 <= i)
+    assert not (12 < i) 
+    assert not (12 <= i)
+
+    assert 12 > i 
+    assert 12 >= i
+    assert 10 > i 
+    assert 10 >= i
+    assert not (8 > i) 
+    assert 8 >= i
+    assert not (0 > i) 
+    assert 0 >= i
+    assert not (-1 > i) 
+    assert not (-1 >= i)
+
+    assert -I.inf < i 
+    assert -I.inf <= i
+    assert not (-I.inf > i) 
+    assert not (-I.inf >= i)
+
+    assert I.inf > i 
+    assert I.inf >= i
+    assert not (I.inf < i) 
+    assert not (I.inf <= i)
 
 
 def test_comparisons_of_unions():
@@ -458,10 +578,9 @@ def test_iteration():
     with pytest.raises(IndexError):
         i1[3]
 
-    assert len(I.empty()) == 0
-    assert list(I.empty()) == []
-    with pytest.raises(IndexError):
-        I.empty()[0]
+    assert len(I.empty()) == 1
+    assert list(I.empty()) == [I.empty().to_atomic()]
+    assert I.empty()[0] == I.empty().to_atomic()
 
 
 def test_complement():
